from ._show import show_moran_scatterplot, show_network, show_voronoi
from ._tendency import show_tendency, show_curves_cluster, show_box_plot
from ._neighborhood import show_neighborhood_analysis, show_infiltration_analysis
from ._ccc import show_ccc_netplot, show_ccc_chordplot, show_ccc_dotplot, show_ccc_embedding
from ._niche import show_niche_environment, show_niche_sample_heatmap, show_celltype_niche_heatmap, \
    show_celltype_sample_heatmap
from ._tensor import show_factor_matrix_in_CP_tensor, show_factor_matrix_in_tucker_tensor, show_proportion_in_tucker, \
    show_proportion_in_CP
