#!/usr/bin/env python
# -*- coding:utf-8 -*-
from . import pp
from . import tl
from . import pl
from .tl import other_package_without_pip

__version__ = "v0.1.7"

"""
# Author: Lihong lab
# File Name:SOAPy
# github:https://github.com/LiHongCSBLab/SOAPy
# Description:
"""

__author__ = "Cancer system biology lab"
__email__ = [
    "wangheqi2021@sinh.ac.cn"
    "lijiarong2020@sinh.ac.cn"
    "lihong01@sibs.ac.cn"
]



